/* 
 *
 * $Id: k3bmonkeydecoder.cpp,v 1.1.1.1 2004/03/09 21:34:16 trueg Exp $
 * Copyright (C) 2003 Sebastian Trueg <trueg@k3b.org>
 *
 * This file is part of the K3b project.
 * Copyright (C) 1998-2003 Sebastian Trueg <trueg@k3b.org>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 * See the file "COPYING" for the exact licensing terms.
 */

#include "k3bmonkeydecoder.h"

#include "All.h"
#include "MACLib.h"
#include "APETag.h"

#include <qcstring.h>
#include <qfile.h>

#include <kdebug.h>
#include <klocale.h>


K3bMonkeyDecoderFactory::K3bMonkeyDecoderFactory( QObject* parent, const char* name )
  : K3bAudioDecoderFactory( parent, name )
{
  s_instance = new KInstance( "k3bmonkeydecoder" );
}


K3bMonkeyDecoderFactory::~K3bMonkeyDecoderFactory()
{
}


K3bPlugin* K3bMonkeyDecoderFactory::createPluginObject( QObject* parent, 
							const char* name,
							const QStringList& )
{
  return new K3bMonkeyDecoder( parent, name );
}


bool K3bMonkeyDecoderFactory::canDecode( const KURL& url )
{
  QCString filename = QFile::encodeName( url.path() );

  int error = 0;

  IAPEDecompress* decoder = CreateIAPEDecompress( filename.data(), &error );
  if( decoder ) {
    delete decoder;
    return true;
  }
  else {
    kdDebug() << "(K3bMonkeyDecoderFactory) failed to open APE file with error " << error << endl;
    return false;
  }
}



class K3bMonkeyDecoder::Private
{
public:
  Private()
    : decoder(0) {
  }

  IAPEDecompress* decoder;
};


K3bMonkeyDecoder::K3bMonkeyDecoder( QObject* parent, const char* name )
  : K3bAudioDecoder( parent, name )
{
  d = new Private();
}


K3bMonkeyDecoder::~K3bMonkeyDecoder()
{
  delete d->decoder;
  delete d;
}


QString K3bMonkeyDecoder::fileType() const
{
  return "Monkey Audio";
}


bool K3bMonkeyDecoder::analyseFileInternal( K3b::Msf& frames, int& samplerate, int& ch )
{
  delete d->decoder;
  int error;
  d->decoder = CreateIAPEDecompress( filename().data(), &error );
  if( d->decoder ) {

    switch( d->decoder->GetInfo(APE_INFO_COMPRESSION_LEVEL) ) {
    case COMPRESSION_LEVEL_FAST: 
      addTechnicalInfo( i18n("Compression level"), i18n("Fast") );
      break;
    case COMPRESSION_LEVEL_NORMAL:
      addTechnicalInfo( i18n("Compression level"), i18n("Normal") );
      break;
    case COMPRESSION_LEVEL_HIGH:
      addTechnicalInfo( i18n("Compression level"), i18n("High") );
      break;
    case COMPRESSION_LEVEL_EXTRA_HIGH:
      addTechnicalInfo( i18n("Compression level"), i18n("Extra High") );
      break;
    }

    addTechnicalInfo( i18n("Sampling Rate"), QString("%1 Hz").arg(d->decoder->GetInfo(APE_INFO_SAMPLE_RATE)) );
    addTechnicalInfo( i18n("Sample Size"), QString("%1 bps").arg(d->decoder->GetInfo(APE_INFO_BITS_PER_SAMPLE)) );
    addTechnicalInfo( i18n("Channels"), QString::number(d->decoder->GetInfo(APE_INFO_CHANNELS)) );
    addTechnicalInfo( i18n("Peak level"), QString::number(d->decoder->GetInfo(APE_INFO_PEAK_LEVEL)) );
    addTechnicalInfo( i18n("Encoder Version"), QString::number( static_cast<double>(d->decoder->GetInfo(APE_INFO_FILE_VERSION)) / 1000.0, 'f', 2 ) );
    
    frames = K3b::Msf::fromSeconds(static_cast<double>(d->decoder->GetInfo(APE_INFO_LENGTH_MS)) / 1000.0);
    samplerate = d->decoder->GetInfo(APE_INFO_SAMPLE_RATE);
    ch = d->decoder->GetInfo(APE_INFO_CHANNELS);

    CAPETag* pAPETag = (CAPETag*)d->decoder->GetInfo(APE_INFO_TAG);
    bool bHasID3Tag = pAPETag->GetHasID3Tag();
    bool bHasAPETag = pAPETag->GetHasAPETag();
    if( bHasID3Tag || bHasAPETag ) {
      // iterate through all the tag fields
      bool bFirst = true;
      CAPETagField* pTagField;
      while( pAPETag->GetNextTagField(bFirst, &pTagField) ) {
	bFirst = false;
	QString name(pTagField->GetFieldName());
	if( name.upper() == "TITLE" )
	  addMetaInfo( META_TITLE, QString::fromUtf8( pTagField->GetFieldValue() ) );
	else if( name.upper() == "ARTIST" )
	  addMetaInfo( META_ARTIST, QString::fromUtf8( pTagField->GetFieldValue() ) );
	else if( name.upper() == "COMMENT" )
	  addMetaInfo( META_COMMENT, QString::fromUtf8( pTagField->GetFieldValue() ) );
	else
	  kdDebug() << "(K3bMonkeyDecoder) unused tag: " << name << endl;
      }
    }

    delete d->decoder;
    d->decoder = 0;

    return true;
  }
  else {
    kdDebug() << "(K3bMonkeyDecoderFactory) failed to open APE file with error " << error << endl;
    return false;
  }
}


bool K3bMonkeyDecoder::initDecoderInternal()
{
  delete d->decoder;
  d->decoder = CreateIAPEDecompress( filename().data() );

  return ( d->decoder != 0 );
}


void K3bMonkeyDecoder::cleanup()
{
  delete d->decoder;
  d->decoder = 0;
}


bool K3bMonkeyDecoder::seekInternal( const K3b::Msf& pos )
{
  return( d->decoder->Seek( pos.totalFrames() * d->decoder->GetInfo(APE_INFO_SAMPLE_RATE) / 75 ) == 0 );
}


int K3bMonkeyDecoder::decodeInternal( char* data, int maxLen )
{
  int nBlocksRetrieved = 0;
  if( d->decoder->GetData( data, maxLen/d->decoder->GetInfo(APE_INFO_BLOCK_ALIGN), &nBlocksRetrieved) == 0 ) {

    nBlocksRetrieved *= d->decoder->GetInfo(APE_INFO_BLOCK_ALIGN);

    // we need to swap the bytes since the APE decoder
    // returnes little endian data
    for( int i = 0; i < nBlocksRetrieved; i+=2 ) {
      char buf = data[i];
      data[i] = data[i+1];
      data[i+1] = buf;
    }

    return nBlocksRetrieved;
  }
  else
    return -1;
}


#include "k3bmonkeydecoder.moc"
